/*
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2011 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


package org.ow2.orchestra.designer.view.processEditor {
import flash.display.DisplayObject;
import flash.events.MouseEvent;
import flash.filters.BitmapFilter;
import flash.filters.BitmapFilterQuality;
import flash.filters.GlowFilter;
import flash.geom.Point;
import flash.geom.Rectangle;

import mx.collections.ArrayCollection;
import mx.events.CollectionEvent;
import mx.events.CollectionEventKind;
import mx.events.FlexEvent;
import mx.states.SetEventHandler;
import mx.states.State;

import org.ow2.orchestra.designer.bpmn.model.AbstractElement;
import org.ow2.orchestra.designer.bpmn.model.AbstractElementWithPosition;
import org.ow2.orchestra.designer.bpmn.views.AbstractElementView;
import org.ow2.orchestra.designer.bpmn.views.IFlowNodeView;
import org.ow2.orchestra.designer.bpmn.views.ProcessPoolView;
import org.ow2.orchestra.designer.events.ProcessSelectionElementCoordinatesChangeEvent;

public class ProcessEditorMovingStateHelper extends ProcessEditorFlowStateHelper {
  private var deltaX:int = 0;
  private var deltaY:int = 0;

  private var elementsMove:ArrayCollection = new ArrayCollection();
  private var elementsInCollision:ArrayCollection = new ArrayCollection();

  private var collisionFilter:BitmapFilter = new GlowFilter(0xCD0000, 1, 12, 12, 2, BitmapFilterQuality.MEDIUM);
  private var selectionFilter:BitmapFilter;

  public function ProcessEditorMovingStateHelper(processEditorView:ProcessEditorView) {
    super(processEditorView);
  }

  /**
   * Generate "moving" state.
   * In the moving state:
   * - on mouse move, the selected elements move to follow the mouse cursor.
   * - on mouse up, this view goes back to the "default" state.
   * - when the mouse leaves this view, the view goes back to the "default" state.
   */
  public function addMovingState():void {
    var state:State = new State();
    state.name = "moving";
    processEditorView.states.push(state);

    // handler for mouse up event
    var mouseUpListener:SetEventHandler = new SetEventHandler(processEditorView, MouseEvent.MOUSE_UP);
    mouseUpListener.handlerFunction = movingMouseUp;
    state.overrides.push(mouseUpListener);

    // handler for mouse move event
    var mouseMoveListener:SetEventHandler = new SetEventHandler(processEditorView, MouseEvent.MOUSE_MOVE);
    mouseMoveListener.handlerFunction = movingMouseMove;
    state.overrides.push(mouseMoveListener);

    state.addEventListener(FlexEvent.ENTER_STATE, movingStateEnter);
    state.addEventListener(FlexEvent.EXIT_STATE, movingStateExit);
  }

  private function movingStateEnter(event:FlexEvent):void {
    // hide context panel
    processEditorView.hideElementContextPanel();

    elementsInCollision.addEventListener(CollectionEvent.COLLECTION_CHANGE, elementsInCollisionChanged);
    this.deltaX = 0;
    this.deltaY = 0;
    elementsMove.removeAll();
    for each (var element:AbstractElement in processEditorView.processEditorModel.selectedElements) {
      if (element is AbstractElementWithPosition) {
        elementsMove.addItem(new ElementCoordinatesUpdate((element as AbstractElementWithPosition)));
      }
    }
    //Retrieve selection filter
    var elementView:AbstractElementView =
            processEditorView.processView.getElementViewByModel((elementsMove.getItemAt(0) as ElementCoordinatesUpdate).elementModel);
    selectionFilter = elementView.borderImage.filters[0];
  }

  private function movingStateExit(event:FlexEvent):void {
    for each (var abstractElementView:AbstractElementView in elementsInCollision) {
      elementsInCollision.removeItemAt(elementsInCollision.getItemIndex(abstractElementView));
    }
    elementsInCollision.removeEventListener(CollectionEvent.COLLECTION_CHANGE, elementsInCollisionChanged);
    // selected element is moved, generate a process modification event.
    new ProcessSelectionElementCoordinatesChangeEvent(
            ProcessSelectionElementCoordinatesChangeEvent.COORDINATES_CHANGE,
            processEditorView.processEditorModel,
            elementsMove
    ).dispatch();
  }

  private function elementsInCollisionChanged(event:CollectionEvent):void {
    var changedElements:Array = event.items;
    switch (event.kind) {
      case CollectionEventKind.ADD:
        for each (var addedElement:AbstractElementView in changedElements) {
          //Update view
          addedElement.borderImage.filters = [collisionFilter];
        }
        break;

      case CollectionEventKind.REMOVE:
        for each (var removedElement:AbstractElementView in changedElements) {
          if (processEditorView.processEditorModel.selectedElements.getItemIndex(removedElement.model) >= 0) {
            removedElement.borderImage.filters = [selectionFilter];
          } else {
            removedElement.borderImage.filters = [];
          }

        }
        break;
    }
  }

  /**
   * handler for mouse up events when this view is in the "moving" state.
   * The view goes back to the "default" state.
   */
  private function movingMouseUp(event:MouseEvent):void {
    // go back to the "default" state.
    processEditorView.currentState = "default";
  }


  /**
   * handler for mouse move event when this view is in the "moving" state.
   * Update selected elements positions to follow the mouse.
   */
  private function movingMouseMove(event:MouseEvent):void {

    // Get mouse event position. The coordinates are relative to the target element
    var targetLocalMousePoint:Point = new Point(event.localX, event.localY);
    // Transform the coordinates to be relative to this view
    var mousePoint:Point = processEditorView.globalToLocal(DisplayObject(event.target).localToGlobal(targetLocalMousePoint));

    var deltaX:int = mousePoint.x - processEditorView.selectionRect.x;
    var deltaY:int = mousePoint.y - processEditorView.selectionRect.y;
    processEditorView.selectionRect.x = mousePoint.x;
    processEditorView.selectionRect.y = mousePoint.y;

    this.deltaX += deltaX;
    this.deltaY += deltaY;
    var collisionObjects:ArrayCollection = new ArrayCollection();
    for each (var elementMoveModel:ElementCoordinatesUpdate in elementsMove) {
      var elementView:AbstractElementView = processEditorView.processView.getElementViewByModel(elementMoveModel.elementModel);
      if (elementView is IFlowNodeView) {
        var viewRect:Rectangle = IFlowNodeView(elementView).getCollisionRectangle(processEditorView);
        viewRect.x += deltaX;
        viewRect.y += deltaY;
        //First try with the current delta
        var precedentColLength:int = collisionObjects.length;
        collisionObjects.addAll(isInCollisionWithAnotherElement(viewRect, elementView));
        if (collisionObjects.length - precedentColLength == 0) {
          if (viewRect.x > ProcessPoolView.POOL_HEADER_WIDTH) {
            elementMoveModel.deltaX += deltaX;
            elementView.x += deltaX;
          }
          if (viewRect.y > processEditorView.processEditorModel.processModel.y) {
            elementMoveModel.deltaY += deltaY;
            elementView.y += deltaY;
          }
        } else {
          // then try with the total delta (jump over feature)
          var initialViewRectangle:Rectangle = IFlowNodeView(elementView).getCollisionRectangle(processEditorView).clone();
          initialViewRectangle.x = elementMoveModel.elementModel.x + this.deltaX;
          initialViewRectangle.y = elementMoveModel.elementModel.y + this.deltaY;
          if (isInCollisionWithAnotherElement(initialViewRectangle, elementView).length == 0) {
            if (initialViewRectangle.x > ProcessPoolView.POOL_HEADER_WIDTH) {
              elementMoveModel.deltaX = this.deltaX;
              elementView.x = initialViewRectangle.x;
            }
            if (initialViewRectangle.y > processEditorView.processEditorModel.processModel.y) {
              elementMoveModel.deltaY = this.deltaY;
              elementView.y = initialViewRectangle.y;
            }
          }
        }
      }
    }
    if (collisionObjects.length == 0) {
      for each (var abstractElementView:AbstractElementView in elementsInCollision) {
        elementsInCollision.removeItemAt(elementsInCollision.getItemIndex(abstractElementView));
      }
    } else {
      elementsInCollision.addAll(collisionObjects);
    }
  }

  private function isInCollisionWithAnotherElement(elementRect:Rectangle, self:AbstractElementView):ArrayCollection {
    var collisionObjects:ArrayCollection = new ArrayCollection();
    for each (var displayObject:DisplayObject in processEditorView.processView.getChildren()) {
      if (displayObject is AbstractElementView
              && displayObject is IFlowNodeView
              && displayObject != self) {
        var targetRect:Rectangle = IFlowNodeView(displayObject).getCollisionRectangle(processEditorView);
        if (elementRect.intersects(targetRect)) {
          collisionObjects.addItem(displayObject);
        }
      }
    }
    return collisionObjects;
  }
}
}
